#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

u"""Pure Python implementation of Map-Reduce over AMQP

"""


#==============================================================================
# CONSTANTS
#==============================================================================

# : This is the project name
PRJ = "Poopy"

# : The project version as tuple of strings
VERSION = ("0", "2", "5")

# : The project version as string
STR_VERSION = ".".join(VERSION)
__version__ = STR_VERSION

DOC = __doc__

# : The short description for pypi
SHORT_DESCRIPTION = []
for line in DOC.splitlines():
    if not line.strip():
        break
    SHORT_DESCRIPTION.append(line)
SHORT_DESCRIPTION = u" ".join(SHORT_DESCRIPTION)
del line


# : Clasifiers for optimize search in pypi
CLASSIFIERS = (
    "Topic :: Utilities",
    "License :: OSI Approved :: BSD License",
    "Programming Language :: Python :: 2",
)

URL = "http://poopy.jbcabral.org"

DOC_URL = ""

AUTHOR = "JuanBC"

# : Email ot the autor
EMAIL = "jbc.develop@gmail.com"

# : The license name
LICENSE = "WISKEY-WARE"

FULL_LICENSE = u""""THE WISKEY-WARE LICENSE":
<jbc.develop@gmail.com> wrote this program. As long as you retain this notice
you can do whatever you want with this stuff. If we meet some day, and you
think this stuff is worth it, you can buy me a WISKEY in return Juan BC
"""

WARRANTY = u"""{}
This program comes with ABSOLUTELY NO WARRANTY; for details see {}
""".format(FULL_LICENSE, URL)



# : Keywords for search of pypi
KEYWORDS = """amqp rabitmq map reduce"""

# : If the program is en debug mode
DEBUG = __debug__


#==============================================================================
# MAIN
#==============================================================================

if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""Implementation of execution node of Poopy

"""


#==============================================================================
# IMPORTS
#==============================================================================

import time
import multiprocessing

from . import connection, conf, serializer


#==============================================================================
# CONSTANTS
#==============================================================================

PONG_E = "pong_exchange"

logger = conf.getLogger("Pong")


#==============================================================================
# CLASS
#==============================================================================

class PongSubscriber(multiprocessing.Process):

    def __init__(self, conn, conf, *args, **kwargs):
        super(PongSubscriber, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf
        self._queue = multiprocessing.Queue()
        self._oldtimes = []
        self._buff = {}

    def reload_buff(self):
        while self._queue.qsize():
            itime, data = self._queue.get_nowait().split("::", 1)
            rconf = conf.conf(**serializer.loads(data))
            self._buff[rconf.UUID] = (float(itime), rconf)

    def uuids(self):
        self.reload_buff()
        return [k for k in self._buff.keys() if self.is_node_alive(k)]

    def is_node_alive(self, uuid):
        self.reload_buff()
        return (
            uuid in self._buff and
            time.time() - self._buff[uuid][0] < self.lconf.TTL
        )

    def get_node_conf(self, uuid):
        if self.is_node_alive(uuid):
            return self._buff[uuid][1]

    # callback
    def _callback(self, ch, method, properties, body):
        if self._queue.qsize() == 0:
            self._oldtimes = []
        for oldtime in tuple(self._oldtimes):
            if time.time() - oldtime > self.lconf.TTL:
                self._queue.get_nowait()
                self._oldtimes.remove(oldtime)
        itime = time.time()
        self._queue.put_nowait("{}::{}".format(itime, body))
        self._oldtimes.append(itime)

    def run(self):
        conn = connection.PoopyConnection(self.conn)
        conn.exchange_consume(PONG_E, self._callback)


class PongPublisher(multiprocessing.Process):

    def __init__(self, conn, conf, *args, **kwargs):
        super(PongPublisher, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf

    def run(self):
        logger.info(
            "Announcing myself with uuid '{}' every '{}' seconds".format(
                self.lconf.UUID, self.lconf.SLEEP
            )
        )

        body = serializer.dumps(self.lconf._asdict())
        conn = connection.PoopyConnection(self.conn)
        channel = conn.channel()
        channel.exchange_declare(exchange=PONG_E, type='fanout')
        while True:
            channel.basic_publish(exchange=PONG_E, routing_key='', body=body)
            conn.sleep(self.lconf.SLEEP)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""Poopy command line interface

"""


#==============================================================================
# IMPORTS
#==============================================================================

import itertools
import os
import runpy
import argparse
import sys
import uuid
import codecs
import traceback
import time
import random
import pickle
import multiprocessing
import contextlib
import datetime

from . import PRJ, STR_VERSION, DOC, WARRANTY
from . import (
    conf, connection, script, pong_node, poopyfs_node,
    script_node, map_node, reduce_node, serializer
)


#==============================================================================
# CONSTANTS
#==============================================================================

logger = conf.getLogger()


#==============================================================================
# ERROR
#==============================================================================

class PoopyError(Exception): pass


#==============================================================================
# FUNCTIONS
#==============================================================================

@contextlib.contextmanager
def proccontext():
    """This context manager store the process to be killed after end
    of execution of the end of the block

    """
    procs = set()
    try:
        yield procs
    finally:
        logger.info("Killing {} processes..".format(len(procs)))
        for proc in procs:
            try:
                proc.terminate()
            except Exception as err:
                logger.warning(err)


def main():
    lconf = conf.conf_from_file()

    parser = argparse.ArgumentParser(
        PRJ, version=STR_VERSION, description=DOC,
        epilog=WARRANTY
    )


    subparsers = parser.add_subparsers(help="Commands help")

    #==========================================================================
    # Create Empty Script
    #==========================================================================

    def manage_createscript(args):
        if os.path.exists(args.filepath):
            raise IOError("File '{}' already exists".format(args.filepath))
        with open(args.filepath, "w") as fp:
            fp.write(script.SCRIPT_TEMPLATE)
        logger.info("Script created at '{}'".format(args.filepath))

    createscript_cmd = subparsers.add_parser(
        'createscript', help='Create a new Poopy script'
    )
    createscript_cmd.add_argument('filepath', help='file to create ')
    createscript_cmd.set_defaults(func=manage_createscript)

    #==========================================================================
    # UPLOAD SUBPARSE
    #==========================================================================

    def manage_upload(args):
        with proccontext() as ctx:
            conn = args.connection
            logger.info("Start discover nodes...")
            pong_sub = pong_node.PongSubscriber(conn, lconf)
            ctx.add(pong_sub)
            pong_sub.start()

            logger.info("Start uploading file...")
            poopyfs_pub = poopyfs_node.PoopyFSPublisher(
                conn, lconf, args.filepath, args.poopyFSpath
            )
            ctx.add(poopyfs_pub)
            poopyfs_pub.start()
            poopyfs_pub.join()


    upload_cmd = subparsers.add_parser('upload', help='upload file to poopyFS')
    upload_cmd.add_argument('connection', help="AMQP URL")
    upload_cmd.add_argument('filepath', help='file to upload')
    upload_cmd.add_argument('poopyFSpath', help='file path to upload')
    upload_cmd.set_defaults(func=manage_upload)

    #==========================================================================
    # DEPLOY SUBPARSE
    #==========================================================================

    def manage_deploy(args):
        with proccontext() as ctx:
            conn = args.connection

            msg = "Starting poopyFS on '{}'..."
            logger.info(msg.format(lconf.POOPY_FS))
            poopyfs_sub = poopyfs_node.PoopyFSSuscriber(conn, lconf)
            ctx.add(poopyfs_sub)
            poopyfs_sub.start()

            msg = "Starting scripts deployment storage on '{}'..."
            logger.info(msg.format(lconf.SCRIPTS))
            script_sub = script_node.ScriptSuscriber(conn, lconf)
            ctx.add(script_sub)
            script_sub.start()

            msg = "Preparing mappers..."
            logger.info(msg)
            map_sub = map_node.MapSubscriber(conn, lconf)
            ctx.add(map_sub)
            map_sub.start()

            msg = "Preparing reducers..."
            logger.info(msg)
            reduce_sub = reduce_node.ReduceSubscriber(conn, lconf)
            ctx.add(reduce_sub)
            reduce_sub.start()

            msg = "Start announce my existence..."
            logger.info(msg)
            pong_pub = pong_node.PongPublisher(conn, lconf)
            ctx.add(pong_pub)
            pong_pub.start()
            pong_pub.join()


    deploy_cmd = subparsers.add_parser('deploy', help='Deploy Poopy node')
    deploy_cmd.add_argument('connection', help="AMQP URL")
    deploy_cmd.set_defaults(func=manage_deploy)

    #==========================================================================
    # RUN SUBPARSE
    #==========================================================================

    def manage_run(args):
        with proccontext() as ctx:
            conn = args.connection
            scriptpath = args.script
            clsname = args.clsname
            out = args.out

            if not os.path.isdir(out):
                os.makedirs(out)
            outpath = os.path.join(
                out, datetime.datetime.now().isoformat().replace(":", "-")
            )

            logger.info("Start discover nodes...")
            pong_sub = pong_node.PongSubscriber(conn, lconf)
            ctx.add(pong_sub)
            pong_sub.start()

            iname = "i{}_{}".format(
                uuid.uuid4().hex, os.path.basename(scriptpath)
            )
            logger.info("Iname generated: {}".format(iname))

            logger.info("Deploying script...")
            script_pub = script_node.ScriptPublisher(
                conn, lconf, scriptpath, iname
            )
            ctx.add(script_pub)
            script_pub.start()
            script_pub.join()

            wait = lconf.SLEEP + lconf.SLEEP / 2.0 + 0.5
            logger.info("Run will start in {} seconds...".format(wait))
            time.sleep(wait)
            uuids = pong_sub.uuids()
            if not uuids:
                msg = "No nodes found. Aborting..."
                raise PoopyError(msg)

            logger.info("Found {} nodes".format(len(uuids)))
            logger.info("Starting Map Responsers...")
            mapr_sub = map_node.MapResultSubscriber(conn, conf, uuids)
            ctx.add(mapr_sub)
            mapr_sub.start()

            logger.info("Starting Reduce Responsers...")
            reducer_sub = reduce_node.ReduceResultSubscriber(conn, conf, uuids)
            ctx.add(reducer_sub)
            reducer_sub.start()

            logger.info("Starting Mapperss...")
            map_pub = map_node.MapPublisher(
                conn, conf, scriptpath, iname, clsname, uuids
            )
            ctx.add(map_pub)
            map_pub.start()
            map_pub.join()

            while not mapr_sub.ended():
                time.sleep(lconf.SLEEP)

            logger.info("Mappers finished")
            results = mapr_sub.results()

            logger.info("Starting Reducers...")
            reduce_pub = reduce_node.ReducePublisher(
                conn, conf, scriptpath, iname, clsname, uuids, results
            )
            ctx.add(reduce_pub)
            reduce_pub.start()
            reduce_pub.join()

            while not reducer_sub.ended():
                time.sleep(lconf.SLEEP)

            logger.info("Reducers finished")
            results = reducer_sub.results()

            logger.info("Writing output")
            with open(outpath, "w") as fp:
                data = serializer.dumps(results)
                fp.write(data)

            logger.info((
                "Your data is pickled in base64 here {}\n"
                "If you want to extract the model use 'serializer.load|s'"
            ).format(outpath))

    run_cmd = subparsers.add_parser('run', help='run script on Poopy cluster')
    run_cmd.add_argument('connection', help="AMQP URL")
    run_cmd.add_argument('script', help='script to run')
    run_cmd.add_argument('clsname', help='class name inside the script')
    run_cmd.add_argument('out', help='output directory')
    run_cmd.set_defaults(func=manage_run)

    args = parser.parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""Script model for Poopy

"""


#==============================================================================
# IMPORTS
#==============================================================================

import abc
import codecs

import unicodecsv as csv

import numpy as np
from scipy.io import arff


#==============================================================================
# BASE
#==============================================================================

class BaseReader(object):

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def load(self, fpath):
        raise NotImplementedError()

    @classmethod
    def subclasses(cls):
        subclss = set()
        def sub_of(actualcls):
            for sc in actualcls.__subclasses__():
                subclss.add(sc)
                sub_of(sc)
        sub_of(cls)
        return tuple(subclss)

    @classmethod
    def reader_by_name(cls, name):
        for sc in cls.subclasses():
            if sc.__name__ == name:
                return sc()
        raise ValueError("Reader whit name '{}' not found".format(name))


#==============================================================================
# ARFF
#==============================================================================

class ARFFReader(BaseReader):

    def load(self, fpath, **kwargs):
        with open(fpath) as fp:
            return arff.loadarff(fp, **kwargs)


class CSVReader(BaseReader):

    def load(self, fpath, **kwargs):
        with open(fpath) as fp:
            reader = csv.reader(fp, **kwargs)
            return iter(reader)


class NumericCsv(BaseReader):

    def load(self, fpath, **kwargs):
        with open(fpath) as fp:
            return np.loadtxt(fp, **kwargs)


class TxtReader(BaseReader):

    def load(self, fp, **kwargs):
        with codecs.open(fpath, **kwargs) as fp:
            return fp.readlines()






#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""Map node of Poopy

"""


#==============================================================================
# IMPORTS
#==============================================================================

import time
import os
import multiprocessing

from . import connection, conf, serializer, script, readers, poopyfs


#==============================================================================
# CONSTANTS
#==============================================================================

MAP_E = "map_exchange"

MAP_RESPONSE_E = "map_response_exchange"

logger = conf.getLogger("Map")


#==============================================================================
# RESULT SUBSCRIBER
#==============================================================================

class MapResultSubscriber(multiprocessing.Process):

    def __init__(self, conn, conf, uuids, *args, **kwargs):
        super(MapResultSubscriber, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf
        self.uuids = set(uuids)
        self._queue = multiprocessing.Queue()
        self._ready = set()
        self._buff = {}

    def results(self):
        return self._buff if self._ready == self.uuids else {}

    def ended(self):
        while self._queue.qsize():
            body = self._queue.get_nowait()
            data = serializer.loads(body)
            self._ready.add(data["uuid"])
            k, v = data["k"], data["v"]
            self._buff.setdefault(k, []).append(v)
        return self._ready == self.uuids

    # callback
    def _callback(self, ch, method, properties, body):
        self._queue.put(body)

    def run(self):
        conn = connection.PoopyConnection(self.conn)
        conn.exchange_consume(MAP_RESPONSE_E, self._callback)


#==============================================================================
# SUBSCRIBER
#==============================================================================

class MapSubscriber(multiprocessing.Process):

    def __init__(self, conn, conf, *args, **kwargs):
        super(MapSubscriber, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf
        self._pconn = None

    # callback
    def _callback(self, ch, method, properties, body):
        logger.info("Map Data recived!")
        data = serializer.loads(body)
        iname = data["iname"]
        clsname = data["clsname"]
        files = data["files"]
        inamepath = os.path.join(self.lconf.SCRIPTS, iname)
        instance = script.cls_from_path(inamepath, clsname)()
        job = script.Job(instance, clsname, iname, self.lconf.UUID)

        logger.info("Preparing for run map for job '{}'".format(job.name))
        conn = connection.PoopyConnection(self.conn)
        channel = conn.channel()
        channel.exchange_declare(exchange=MAP_RESPONSE_E, type='fanout')

        def emiter(k, v):
            body = serializer.dumps({"k": k, "v": v, "uuid": self.lconf.UUID})
            channel.basic_publish(
                exchange=MAP_RESPONSE_E, routing_key='', body=body
            )

        context = script.Context(emiter, job)
        try:
            for fpath, reader_name, kwargs in files:
                logger.info("Running map for '{}'".format(fpath))
                reader = readers.BaseReader.reader_by_name(reader_name)
                lpath = poopyfs.resolve_poopyfslocal(
                    fpath, self.lconf, clean=True
                )
                data = reader.load(lpath, **kwargs)
                instance.map(fpath, data, context)
        except Exception as err:
            logger.error(str(err))

    def run(self):
        conn = connection.PoopyConnection(self.conn)
        conn.exchange_consume(MAP_E, self._callback, self.lconf.UUID)


#==============================================================================
# PUBLISHER
#==============================================================================

class MapPublisher(multiprocessing.Process):

    def __init__(self, conn, conf, script,
                 iname, clsname, uuids, *args, **kwargs):
        super(MapPublisher, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf
        self.script = script
        self.iname = iname
        self.clsname = clsname
        self.uuids = uuids

    def split_files(self, job):
        files = job.input_path
        chunk_size = len(self.uuids)
        for idx, chunk_from in enumerate(range(0, len(files), chunk_size)):
            chunk_to = chunk_from + chunk_size
            chunk = []
            for filename in files[chunk_from:chunk_to]:
                if isinstance(filename, (tuple, list)) and len(filename)  == 3:
                    filename, parser, kwargs = filename
                    chunk.append((filename, parser.__name__, kwargs))
                elif isinstance(filename, (tuple, list)):
                    filename, parser, = filename
                    chunk.append((filename, parser.__name__, {}))
                elif isinstance(filename, basestring):
                    chunk.append((filename, readers.TxtReader.__name__, {}))
                else:
                    raise ValueError(
                        "Filename must be an string or an tuple or "
                        "list with 2 or 3 elemenets"
                    )
            yield self.uuids[idx], chunk

    def run(self):
        logger.info("Reading script {}...".format(self.script))
        Cls = script.cls_from_path(self.script, self.clsname)
        instance = Cls()

        logger.info("Reading {} configuration...".format(self.script))
        job = script.Job(instance, self.clsname, self.iname)

        # publishing
        conn = connection.PoopyConnection(self.conn)
        channel = conn.channel()
        channel.exchange_declare(exchange=MAP_E, type='fanout')

        for uuid, files in self.split_files(job):
            msg = (
                "Sending Map Execution signal for script '{}' to node '{}'"
            ).format(job.name, uuid)
            logger.info(msg)
            body = serializer.dumps({
                "iname": self.iname, "clsname": self.clsname, "files": files
            })
            channel.basic_publish(exchange=MAP_E, routing_key=uuid, body=body)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""Connection base class

"""


#==============================================================================
# IMPORTS
#==============================================================================

import pika


#==============================================================================
# CLASS
#==============================================================================

class PoopyConnection(pika.BlockingConnection):

    def __init__(self, conn_str, *args, **kwargs):
        self.conn_str = conn_str
        self.params = pika.URLParameters(conn_str)
        super(PoopyConnection, self).__init__(self.params, *args, **kwargs)

    def exchange_consume(self, exchange, callback, routing_key=None):
        channel = self.channel()
        channel.exchange_declare(exchange=exchange, type='fanout')
        result = channel.queue_declare(exclusive=True)
        queue_name = result.method.queue
        channel.queue_bind(
            exchange=exchange, queue=queue_name, routing_key=routing_key
        )
        channel.basic_consume(callback, queue=queue_name, no_ack=False)
        channel.start_consuming()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""This module copy and retrieves the files over a poopFS distributed file
system

"""


#==============================================================================
# IMPORTS
#==============================================================================

import os
import runpy
import uuid
import codecs
import multiprocessing

try:
    import cPickle as pickle
except ImportError:
    import pickle

import pika

from . import connection, serializer, conf, poopyfs


#==============================================================================
# CONSTANTS
#==============================================================================

POOPY_FS_E = "poopyFS_exchange"

logger = conf.getLogger("poopyFS")


#==============================================================================
# CLASS
#==============================================================================

class PoopyFSSuscriber(multiprocessing.Process):

    def __init__(self, conn, conf, *args, **kwargs):
        super(PoopyFSSuscriber, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf
        if not os.path.isdir(conf.POOPY_FS):
            os.makedirs(conf.POOPY_FS)

    def _callback(self, ch, method, properties, body):
        data = serializer.loads(body)
        fname = data["poopyFSpath"]
        src = data["src"]
        logger.info("Receiving {}{}".format(poopyfs.PREFIX_POOPY_FS, fname))
        fpath = poopyfs.resolve_poopyfslocal(fname, self.lconf, clean=False)
        dirname = os.path.dirname(fpath)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        with open(fpath, "w") as fp:
            fp.write(src)

    def run(self):
        conn = connection.PoopyConnection(self.conn)
        conn.exchange_consume(POOPY_FS_E, self._callback)


class PoopyFSPublisher(multiprocessing.Process):

    def __init__(self, conn, conf, filepath, poopyFSpath, *args, **kwargs):
        super(PoopyFSPublisher, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf
        self.filepath = filepath
        self.poopyFSpath = poopyFSpath

    def run(self):
        logger.info(
            "Upoloading '{}' to '{}'".format(self.filepath, self.poopyFSpath)
        )
        to_path = poopyfs.clean_poopyfs_filename(self.poopyFSpath)
        with open(self.filepath, "rb") as fp:
            src = fp.read()
        body = serializer.dumps({"poopyFSpath": to_path, "src": src})

        conn = connection.PoopyConnection(self.conn)
        channel = conn.channel()
        channel.exchange_declare(exchange=POOPY_FS_E, type='fanout')
        channel.basic_publish(exchange=POOPY_FS_E, routing_key='', body=body)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""Distribution for scripts on Poopy

"""


#==============================================================================
# IMPORTS
#==============================================================================

import os
import runpy
import uuid
import codecs
import multiprocessing

try:
    import cPickle as pickle
except ImportError:
    import pickle

import pika

from . import connection, serializer, conf


#==============================================================================
# CONSTANTS
#==============================================================================

SCRIPT_E = "script_exchange"

logger = conf.getLogger("Scripts")


#==============================================================================
# CLASS
#==============================================================================

class ScriptSuscriber(multiprocessing.Process):

    def __init__(self, conn, conf, *args, **kwargs):
        super(ScriptSuscriber, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf
        if not os.path.isdir(conf.SCRIPTS):
            os.makedirs(conf.SCRIPTS)

    def _callback(self, ch, method, properties, body):
        data = serializer.loads(body)
        fname = data["filename"]
        ifname = data["ifilename"]
        src = data["src"]
        logger.info("Receiving {} ({})".format(fname, ifname))
        fpath = os.path.join(self.lconf.SCRIPTS, ifname)
        with open(fpath, "w") as fp:
            fp.write(src)

    def run(self):
        conn = connection.PoopyConnection(self.conn)
        conn.exchange_consume(SCRIPT_E, self._callback)


class ScriptPublisher(multiprocessing.Process):

    def __init__(self, conn, conf, filepath, ifilename, *args, **kwargs):
        super(ScriptPublisher, self).__init__(*args, **kwargs)
        self.conn = conn
        self.lconf = conf
        self.filepath = filepath
        self.ifilename = ifilename

    def run(self):
        logger.info("Deploying script '{}'".format(self.filepath))

        with open(self.filepath, "rb") as fp:
            src = fp.read()
        body = serializer.dumps({
            "filename": os.path.basename(self.filepath),
            "ifilename": self.ifilename, "src": src
        })
        conn = connection.PoopyConnection(self.conn)
        channel = conn.channel()
        channel.exchange_declare(exchange=SCRIPT_E, type='fanout')
        channel.basic_publish(exchange=SCRIPT_E, routing_key='', body=body)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""This file is for distribute Poopy with distutils

"""

#==============================================================================
# CONSTANTS
#==============================================================================

REQUIREMENTS = [
    "pika>=0.9.14",
    "numpy>=1.9.1",
    "scipy>=0.14.0",
    "scikit-learn>=0.15.2",
    "unicodecsv>=0.9.4"
]


#==============================================================================
# FUNCTIONS
#==============================================================================

if __name__ == "__main__":
    import os
    import sys

    from ez_setup import use_setuptools
    use_setuptools()

    from setuptools import setup, find_packages

    import poopy

    setup(
        name=poopy.PRJ.lower(),
        version=poopy.STR_VERSION,
        description=poopy.SHORT_DESCRIPTION,
        author=poopy.AUTHOR,
        author_email=poopy.EMAIL,
        url=poopy.URL,
        license=poopy.LICENSE,
        keywords=poopy.KEYWORDS,
        classifiers=poopy.CLASSIFIERS,
        packages=[pkg for pkg in find_packages() if pkg.startswith("poopy")],
        include_package_data=True,
        py_modules=["ez_setup"],
        entry_points={'console_scripts': ['poopy = poopy.cli:main']},
        install_requires=REQUIREMENTS,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# "THE WISKEY-WARE LICENSE":
# <jbc.develop@gmail.com> wrote this file. As long as you retain this notice
# you can do whatever you want with this stuff. If we meet some day, and you
# think this stuff is worth it, you can buy me a WISKEY in return Juan BC


#==============================================================================
# DOCS
#==============================================================================

"""Script model for Poopy

"""


#==============================================================================
# IMPORTS
#==============================================================================

import abc
import runpy
import inspect
import collections
import copy

from . import readers


#==============================================================================
# CONSTANTS
#==============================================================================

SCRIPT_TEMPLATE = """
#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ampoopq import script


class Script(script.ScriptBase):

    def map(self, k, v, ctx):
        raise NotImplementedError()

    def reduce(self, k, v, ctx):
        for vi in v:
            ctx.emit(k, vi)

    def setup(self, job):
        job.name = "NO-NAME"


""".strip()


#==============================================================================
# SCRIPT CLASSES
#==============================================================================

class ScriptMeta(abc.ABCMeta):

    def __init__(self, *args, **kwargs):
        super(ScriptMeta, self).__init__(*args, **kwargs)

        # modify readers list
        readers_dict = {r.__name__: r for r in readers.BaseReader.subclasses()}
        ReadersClass = collections.namedtuple("Readers", readers_dict.keys())
        self.readers = ReadersClass(**readers_dict)

    def __repr__(cls):
        return (
            getattr(cls, 'class_name', None) or
            super(ScriptMeta, cls).__repr__()
        )


class ScriptBase(object):

    class_name = None
    __metaclass__ = ScriptMeta

    @abc.abstractmethod
    def map(self, k, v, ctx):
        raise NotImplementedError()

    def reduce(self, k, v, ctx):
        for vi in v:
            ctx.emit(k, vi)

    @abc.abstractmethod
    def setup(self, ctx):
        raise NotImplementedError()


#==============================================================================
# JOB
#==============================================================================

class Job(object):

    def __init__(self, script, clsname, iname, uuid=None):
        self._script = script
        self._clsname = clsname
        self._name = "<NO-NAME>"
        self._global_vars = {}
        self._mappers = []
        self._reducers = []
        self._input_path = []
        self._output_path = "<NO-PATH>"
        self._iname = iname

        # populate this conf with the script
        self.script.setup(self)

    @property
    def script(self):
        return self._script

    @property
    def clsname(self):
        return self._clsname

    @property
    def iname(self):
        return self._iname

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, v):
        self._name = unicode(v)

    @property
    def output_path(self):
        return self._output_path

    @output_path.setter
    def output_path(self, v):
        self._output_path = unicode(v)

    @property
    def input_path(self):
        return self._input_path

    @property
    def reducers(self):
        return self._reducers

    @property
    def mappers(self):
        return self._mappers

    @property
    def global_vars(self):
        return self._global_vars

    @property
    def local_vars(self):
        return self._local_vars


#==============================================================================
# CONTEXTS
#==============================================================================

class Context(object):

    def __init__(self, emiter, job):
        self.job = job
        self._emiter = emiter

    def emit(self, k, v):
        self._emiter(k, v)


#==============================================================================
# FUNCTIONS
#==============================================================================

def cls_from_path(path, clsname):
    Cls = runpy.run_path(path)[clsname]
    if not inspect.isclass(Cls) or not issubclass(Cls, ScriptBase):
        msg = "'{}' is not subclass of poopy.script.ScriptBase".format(Cls)
        raise TypeError(msg)
    Cls.class_name = "<class '{}:{}'>".format(path, clsname)
    return Cls








